@@ -1,7 +1,13 @@
 import struct
 import ctypes
-from typing import List, Set, Dict, Tuple
 import capstone as cp
+from typing import (
+	List,
+	Set,
+	Dict,
+	Tuple,
+	Generator
+)
 
 from DyldExtractor.extraction_context import ExtractionContext
 from DyldExtractor.converter import (
@@ -46,12 +52,12 @@ def disasm_lite_new(self, code, offset, count=0, codeOffset=0):
 	# Pass a bytearray by reference
 	if isinstance(code, bytearray):
 		code = ctypes.byref(ctypes.c_char.from_buffer(code, codeOffset))
-	res = cp._cs.cs_disasm(self.csh, code, size, offset, count, ctypes.byref(all_insn))
+	res = cp._cs.cs_disasm(self.csh, code, size, offset, count, ctypes.byref(all_insn))  # noqa
 	if res > 0:
 		try:
 			for i in range(res):
 				insn = all_insn[i]
-				yield (insn.address, insn.size, insn.mnemonic.decode('ascii'), insn.op_str.decode('ascii'))
+				yield (insn.address, insn.size, insn.mnemonic.decode('ascii'), insn.op_str.decode('ascii'))  # noqa
 		finally:
 			cp._cs.cs_free(all_insn, res)
 	else:
@@ -70,236 +76,6 @@ class _ObjCFixerError(Exception):
 
 
 class _ObjCSelectorFixer(object):
-	def __init__(
-		self,
-		extractionCtx: ExtractionContext,
-		delegate: "_ObjCFixer"
-	) -> None:
-		"""Un-does direct selector loading.
-
-		Args:
-			extractionCtx: The extraction context
-			delegate: The delegate to add more data if needed.
-		"""
-
-		super().__init__()
-
-		self._dyldCtx = extractionCtx.dyldCtx
-		self._machoCtx = extractionCtx.machoCtx
-		self._statusBar = extractionCtx.statusBar
-		self._logger = extractionCtx.logger
-		self._delegate = delegate
-
-		self.opStrTrans = str.maketrans("", "", "[]!")
-		pass
-
-	def run(self):
-		try:
-			textSect = self._machoCtx.segments[b"__TEXT"].sects[b"__text"]
-			pass
-		except KeyError:
-			self._logger.error("Unable to get __text section")
-			pass
-
-		self._statusBar.update(status="Fixing Selectors")
-
-		disassembler = cp.Cs(cp.CS_ARCH_ARM64, cp.CS_MODE_LITTLE_ENDIAN)
-
-		textSectAddr = textSect.addr
-		textSectOff = self._dyldCtx.convertAddr(textSectAddr)
-		textSectEnd = textSectOff + textSect.size
-
-		for sectOff in range(0, textSect.size, 4):
-			# Direct Selectors start with an ADRP instruction
-			adrpOff = textSectOff + sectOff
-			if self._dyldCtx.file[adrpOff + 3] & 0x9F != 0x90:
-				continue
-			adrpAddr = textSectAddr + sectOff
-
-			# try to find the matching add instruction
-			addRelOff = self._findAddInstr(
-				disassembler,
-				adrpAddr,
-				adrpOff,
-				textSectEnd
-			)
-			if addRelOff is None:
-				continue
-
-			addOff = adrpOff + addRelOff
-
-			adrpInstr = self._dyldCtx.readFormat(adrpOff, "<I")[0]
-			addInstr = self._dyldCtx.readFormat(addOff, "<I")[0]
-
-			# Find the target of the direct selector load
-			# ADRP
-			immlo = (adrpInstr & 0x60000000) >> 29
-			immhi = (adrpInstr & 0xFFFFE0) >> 3
-			imm = (immhi | immlo) << 12
-			imm = stub_fixer.Arm64Utilities.signExtend(imm, 33)
-
-			adrpResult = (adrpAddr & ~0xFFF) + imm
-
-			# ADD
-			imm = (addInstr & 0x3FFC00) >> 10
-			loadTarget = adrpResult + imm
-
-			# check if it needs fixing
-			if self._machoCtx.containsAddr(loadTarget):
-				continue
-
-			if loadTarget not in self._delegate._selRefCache:
-				# check if its a valid address
-				if not self._dyldCtx.convertAddr(loadTarget):
-					continue
-
-				self._addStringAndFixLoad(
-					loadTarget,
-					adrpInstr,
-					adrpAddr,
-					adrpOff,
-					addInstr,
-					addOff
-				)
-				self._statusBar.update(status="Fixing Selectors")
-				continue
-
-			self._fixLoad(
-				loadTarget,
-				adrpInstr,
-				adrpAddr,
-				adrpOff,
-				addInstr,
-				addOff
-			)
-			self._statusBar.update(status="Fixing Selectors")
-			pass
-		pass
-
-	def _fixLoad(
-		self,
-		loadTarget: int,
-		adrpInstr: int,
-		adrpAddr: int,
-		adrpOff: int,
-		addInstr: int,
-		addOff: int
-	):
-		selRefPtr = self._delegate._selRefCache[loadTarget]
-
-		adrpDelta = (selRefPtr & -4096) - (adrpAddr & -4096)
-		immhi = (adrpDelta >> 9) & (0x00FFFFE0)
-		immlo = (adrpDelta << 17) & (0x60000000)
-		newAdrp = (0x90000000) | immlo | immhi | adrpInstr & 0x1F
-
-		ldrTargetOff = selRefPtr - (selRefPtr & -4096)
-		imm12 = (ldrTargetOff << 7) & 0x3FFC00
-		ldrRegisters = addInstr & 0x3FF
-		newLdr = 0xF9400000 | imm12 | ldrRegisters
-
-		self._machoCtx.writeBytes(adrpOff, struct.pack("<I", newAdrp))
-		self._machoCtx.writeBytes(addOff, struct.pack("<I", newLdr))
-		pass
-
-	def _addStringAndFixLoad(
-		self,
-		loadTarget: int,
-		adrpInstr: int,
-		adrpAddr: int,
-		adrpOff: int,
-		addInstr: int,
-		addOff: int
-	):
-		# There are some files that access strings that do not have
-		# a selector reference. Pull in the string and repoint the
-		# ADRP and ADD to it.
-		stringAddr = self._delegate._processString(loadTarget)
-
-		adrpDelta = (stringAddr & -4096) - (adrpAddr & -4096)
-		immhi = (adrpDelta >> 9) & (0x00FFFFE0)
-		immlo = (adrpDelta << 17) & (0x60000000)
-		newAdrp = (0x90000000) | immlo | immhi | adrpInstr & 0x1F
-
-		addTargetOff = stringAddr - (stringAddr & -4096)
-		imm12 = (addTargetOff << 10) & 0x3FFC00
-		addRegisters = addInstr & 0x3FF
-		newAdd = 0x91000000 | imm12 | addRegisters
-
-		self._machoCtx.writeBytes(adrpOff, struct.pack("<I", newAdrp))
-		self._machoCtx.writeBytes(addOff, struct.pack("<I", newAdd))
-		pass
-
-	def _findAddInstr(
-		self,
-		disassembler: cp.Cs,
-		adrpAddr: int,
-		adrpOff: int,
-		textEnd: int
-	) -> int:
-		"""Find the address of a matching add instruction.
-
-		Args:
-			disassembler: The initalized disassembler.
-			adrpAddr: The address of the ADRP instruction.
-			adrpOff: The file offset of the ADRP instruction.
-			textEnd: The file offset to the end of the text section.
-
-		Return:
-			The relative offset to the add instruction, or None if it
-			could not be found.
-		"""
-
-		# get the destination register for the ADRP
-		# process instructions in 64 byte chunks
-		blockSlice = slice(adrpOff, min(adrpOff + 64, textEnd))
-		instrData = self._dyldCtx.file[blockSlice]
-		blockDisasm = disassembler.disasm_lite(instrData, adrpAddr)
-		adrpInstr = next(blockDisasm)
-
-		if adrpInstr[2] != "adrp":
-			self._logger.error(f"Expected an adrp instruction at {hex(adrpAddr)}")
-			return None
-
-		adrpDestReg = self._getOpcodes(adrpInstr[3])[0]
-
-		instrBlockOff = adrpOff + 4
-		instrBlockAddr = adrpAddr + 4
-		while instrBlockOff < textEnd:
-			for instruction in blockDisasm:
-				# instruction = (address, size, mnemonic, op_str)
-				opcodes = self._getOpcodes(instruction[3])
-
-				# check if the ADRP dest reg matches the base reg for ADD
-				if instruction[2] == "add" and opcodes[1] == adrpDestReg:
-					return instruction[0] - adrpAddr
-
-				# If we find an instruction modifying the register,
-				# it probably isn't a selector reference
-				if adrpDestReg == opcodes[0]:
-					return None
-				pass
-
-			# process the next block if needed
-			instrBlockOff += len(instrData)
-			instrBlockAddr += len(instrData)
-
-			blockSlice = slice(instrBlockOff, min(instrBlockOff + 64, textEnd))
-			instrData = self._dyldCtx.file[blockSlice]
-			blockDisasm = disassembler.disasm_lite(instrData, instrBlockAddr)
-			pass
-
-		return None
-
-	def _getOpcodes(self, opStr: str) -> List[str]:
-		return [
-			opcode.strip()
-			for opcode
-			in opStr.translate(self.opStrTrans).split(",")
-		]
-	pass
-
-
-class _ObjCSelectorFixerV2(object):
 	def __init__(
 		self,
 		extractionCtx: ExtractionContext,
@@ -342,7 +118,10 @@ class _ObjCSelectorFixerV2(object):
 
 		# enumerate the text
 		textSectAddr = textSect.addr
-		textSectOff = self._dyldCtx.convertAddr(textSectAddr)
+		textSectOff, ctx = self._dyldCtx.convertAddr(textSectAddr)
+
+		textFile = ctx.fileCtx
+		writableTextFile = self._machoCtx.fileForAddr(textSectAddr)
 
 		for i, instrData in enumerate(self._textInstr):
 			if instrData[1] != "adrp":
@@ -356,7 +135,7 @@ class _ObjCSelectorFixerV2(object):
 
 			adrpAddr = textSectAddr + (i * 4)
 			adrpOff = textSectOff + (i * 4)
-			adrpInstr = self._dyldCtx.readFormat(adrpOff, "<I")[0]
+			adrpInstr = textFile.readFormat("<I", adrpOff)[0]
 
 			# Find the ADRP result
 			immlo = (adrpInstr & 0x60000000) >> 29
@@ -369,7 +148,7 @@ class _ObjCSelectorFixerV2(object):
 
 			for addInstrIdx in addInstrIdxs:
 				addOff = textSectOff + (addInstrIdx * 4)
-				addInstr = self._dyldCtx.readFormat(addOff, "<I")[0]
+				addInstr = textFile.readFormat("<I", addOff)[0]
 
 				# Test for a special ADD cases
 				if addInstr & 0xffc00000 != 0x91000000:
@@ -395,7 +174,7 @@ class _ObjCSelectorFixerV2(object):
 					immhi = (adrpDelta >> 9) & (0x00FFFFE0)
 					immlo = (adrpDelta << 17) & (0x60000000)
 					newAdrp = (0x90000000) | immlo | immhi | adrpInstr & 0x1F
-					self._machoCtx.writeBytes(adrpOff, struct.pack("<I", newAdrp))
+					writableTextFile.writeBytes(adrpOff, struct.pack("<I", newAdrp))
 					pass
 				else:
 					# Make sure the new address is reachable with the new adrp
@@ -410,7 +189,7 @@ class _ObjCSelectorFixerV2(object):
 				imm12 = (ldrTargetOff << 7) & 0x3FFC00
 				ldrRegisters = addInstr & 0x3FF
 				newLdr = 0xF9400000 | imm12 | ldrRegisters
-				self._machoCtx.writeBytes(addOff, struct.pack("<I", newLdr))
+				writableTextFile.writeBytes(addOff, struct.pack("<I", newLdr))
 
 				self._statusBar.update(status="Fixing Selectors")
 				pass
@@ -423,8 +202,8 @@ class _ObjCSelectorFixerV2(object):
 		self._statusBar.update(status="Disassembling Text (will appear frozen)")
 
 		textSect = self._machoCtx.segments[b"__TEXT"].sects[b"__text"]
-		textSectOff = self._dyldCtx.convertAddr(textSect.addr)
-		textData = bytearray(self._dyldCtx.getBytes(textSectOff, textSect.size))
+		textSectOff, ctx = self._dyldCtx.convertAddr(textSect.addr)
+		textData = bytearray(ctx.fileCtx.getBytes(textSectOff, textSect.size))
 
 		opStrTrans = str.maketrans("", "", "[]!")
 		disassembler = cp.Cs(cp.CS_ARCH_ARM64, cp.CS_MODE_LITTLE_ENDIAN)
@@ -479,11 +258,10 @@ class _ObjCSelectorFixerV2(object):
 		self,
 		startIdx: int,
 		adrpReg: str,
-		_processed: Set[int] = None
 	) -> Set[int]:
 		"""Find ADD instructions given an ADRP register.
 
-		This will recursively follow branches and stop
+		This will iteratively follow branches and stop
 		when the ADRP range ends.
 
 		Args:
@@ -492,98 +270,92 @@ class _ObjCSelectorFixerV2(object):
 			A list of indices to the ADD instructions.
 		"""
 
-		# Keep track of start indexes that are already processed
-		if _processed is None:
-			_processed = {startIdx}
-			pass
-		else:
-			if startIdx in _processed:
-				return set()
-			else:
-				_processed.add(startIdx)
-			pass
-
 		addIdxs = set()
-		i = startIdx
 
-		while i < len(self._textInstr) and i >= 0:
-			address, mnemonic, opcodes = self._textInstr[i]
+		# set of indexes that are or being processed
+		processedIdx = set()
 
-			# check if the ADRP dest reg matches the base reg for the ADD
-			if mnemonic == "add" and opcodes[1] == adrpReg:
-				addIdxs.add(i)
-				pass
+		# list of indices that need to be processed
+		startIndices = [startIdx]
 
-			# If there is an unconditional branch, and it points
-			# within the text section, follow it. If it does not
-			# point within the text section, end the ADRP range.
-			if mnemonic == "b":
-				branchAddr = int(opcodes[0][1:], 16)
-				idxDelta = int((branchAddr - address) / 4)
-				i += idxDelta
-
-				if i in _processed:
-					break
-				else:
-					_processed.add(i)
-
-				if i < 0 or i >= len(self._textInstr):
-					break
+		while len(startIndices):
+			i = startIndices.pop()
+			if i in processedIdx:
 				continue
-
-			# If there is a conditional branch, follow it and continue
-			if mnemonic[0:2] == "b.":
-				branchAddr = int(opcodes[0][1:], 16)
-				idxDelta = int((branchAddr - address) / 4)
-				addIdxs.update(
-					self._findAddInstructions(i + idxDelta, adrpReg, _processed=_processed)
-				)
-				pass
-			elif mnemonic == "cbz" or mnemonic == "cbnz":
-				branchAddr = int(opcodes[1][1:], 16)
-				idxDelta = int((branchAddr - address) / 4)
-				addIdxs.update(
-					self._findAddInstructions(i + idxDelta, adrpReg, _processed=_processed)
-				)
-				pass
-			elif mnemonic == "tbz" or mnemonic == "tbnz":
-				branchAddr = int(opcodes[2][1:], 16)
-				idxDelta = int((branchAddr - address) / 4)
-				addIdxs.update(
-					self._findAddInstructions(i + idxDelta, adrpReg, _processed=_processed)
-				)
+			else:
+				processedIdx.add(i)
 				pass
 
-			# End the ADRP range if the function returns
-			if mnemonic in (
-				"ret",
-				"retaa",
-				"retab"
-			):
-				break
+			while i < len(self._textInstr) and i >= 0:
+				address, mnemonic, opcodes = self._textInstr[i]
 
-			# If we find an instruction modifying the register,
-			# the adrp range probably ended.
-			if adrpReg == opcodes[0]:
-				break
+				# check if the ADRP dest reg matches the base reg for the ADD
+				if mnemonic == "add" and opcodes[1] == adrpReg:
+					addIdxs.add(i)
+					pass
 
-			# These instructions modify 2 registers.
-			if mnemonic in (
-				"ldaxp",
-				"ldnp",
-				"ldpsw",
-				"ldxp",
-				"stlxp",
-				"stnp",
-				"stp",
-				"stxp",
-				"ldp"
-			):
-				if adrpReg == opcodes[1]:
+				# If there is an unconditional branch, and it points
+				# within the text section, follow it. If it does not
+				# point within the text section, end the ADRP range.
+				if mnemonic == "b":
+					branchAddr = int(opcodes[0][1:], 16)
+					idxDelta = int((branchAddr - address) / 4)
+					i += idxDelta
+
+					if i < 0 or i >= len(self._textInstr):
+						break
+
+					startIndices.append(i)
 					break
-				pass
 
-			i += 1
+				# If there is a conditional branch, follow it and continue
+				elif mnemonic[0:2] == "b.":
+					branchAddr = int(opcodes[0][1:], 16)
+					idxDelta = int((branchAddr - address) / 4)
+					startIndices.append(i + idxDelta)
+					pass
+				elif mnemonic == "cbz" or mnemonic == "cbnz":
+					branchAddr = int(opcodes[1][1:], 16)
+					idxDelta = int((branchAddr - address) / 4)
+					startIndices.append(i + idxDelta)
+					pass
+				elif mnemonic == "tbz" or mnemonic == "tbnz":
+					branchAddr = int(opcodes[2][1:], 16)
+					idxDelta = int((branchAddr - address) / 4)
+					startIndices.append(i + idxDelta)
+					pass
+
+				# End the ADRP range if the function returns
+				if mnemonic in (
+					"ret",
+					"retaa",
+					"retab"
+				):
+					break
+
+				# If we find an instruction modifying the register,
+				# the adrp range probably ended.
+				if adrpReg == opcodes[0]:
+					break
+
+				# These instructions modify 2 registers.
+				if mnemonic in (
+					"ldaxp",
+					"ldnp",
+					"ldpsw",
+					"ldxp",
+					"stlxp",
+					"stnp",
+					"stp",
+					"stxp",
+					"ldp"
+				):
+					if adrpReg == opcodes[1]:
+						break
+					pass
+
+				i += 1
+				pass
 			pass
 
 		return addIdxs
@@ -606,25 +378,46 @@ class _ObjCFixer(object):
 	def run(self):
 		# check if the optimization flag is set
 		imageInfo = None
+		imageInfoFile = None
 		for seg in self._machoCtx.segmentsI:
 			if b"__objc_imageinfo" in seg.sects:
 				imageInfo = seg.sects[b"__objc_imageinfo"]
+				imageInfoFile = self._machoCtx.fileForAddr(seg.seg.vmaddr)
 				break
 			pass
 
 		if not imageInfo:
 			return
 
-		flags = self._machoCtx.readFormat(
-			self._dyldCtx.convertAddr(imageInfo.addr) + 4,
-			"<I"
+		flagsOff = self._dyldCtx.convertAddr(imageInfo.addr)[0]
+		flags = imageInfoFile.readFormat(
+			"<I",
+			flagsOff + 4,
 		)[0]
 		if not flags & 0x8:
 			self._logger.info("ObjC was not optimized by Dyld, not fixing ObjC.")
 			return
 
+		# Removed the optimized objc bit
+		flags &= 0xfffffff7
+		imageInfoFile.writeBytes(flagsOff, struct.pack("<I", flags))
+
 		self._createExtraSegment()
 
+		# Get the libobjc.A.dylib image
+		for image in self._dyldCtx.images:
+			path = self._dyldCtx.fileCtx.readString(image.pathFileOffset)
+			if b"libobjc.A.dylib" in path:
+				offset, ctx = self._dyldCtx.convertAddr(image.address)
+				self._libobjcImage = MachOContext(ctx.fileCtx, offset)
+				break
+			pass
+		else:
+			self._logger.error("Unable to find libobjc.A.dylib")
+			return
+
+		self._checkMethodNameStorage()
+
 		# caches that map the original definition address
 		# to its new processed address.
 		self._categoryCache: Dict[int, int] = {}
@@ -637,6 +430,7 @@ class _ObjCFixer(object):
 		self._methodListCache: Dict[int, int] = {}
 		self._stringCache: Dict[int, int] = {}
 		self._intCache: Dict[int, int] = {}
+		self._methodNameCache: Dict[int, int] = {}
 
 		# connects a selrefs old target to its pointer address
 		self._selRefCache: Dict[int, int] = {}
@@ -653,14 +447,117 @@ class _ObjCFixer(object):
 		self._processSections()
 		self._finalizeFutureClasses()
 
-		# self._fixSelectors_OLD()
-		# _ObjCSelectorFixer(self._extractionCtx, self).run()
-		_ObjCSelectorFixerV2(self._extractionCtx, self).run()
+		_ObjCSelectorFixer(self._extractionCtx, self).run()
 
 		self._checkSpaceConstraints()
 		self._addExtraDataSeg()
 		pass
 
+	def _checkMethodNameStorage(self) -> None:
+		"""Check where method names are stored.
+
+		Starting around iOS 15, relative method names
+		pointers are relative to the start of the __OBJC_RO of
+		libobjc, instead of being relative to itself.
+		This tries to detect which is being used.
+		"""
+
+		self._usesObjcRoRelativeNames = False
+		self._optMethodNamesAddr = None
+
+		# Get __objc_scoffs, __objc_classlist
+		objcScoffs = None
+		objcClasslist = None
+		for seg in self._libobjcImage.segmentsI:
+			for sect in seg.sectsI:
+				if sect.sectname == b"__objc_scoffs":
+					objcScoffs = sect
+					pass
+				elif sect.sectname == b"__objc_classlist":
+					objcClasslist = sect
+					pass
+				pass
+
+			if objcScoffs and objcClasslist:
+				break
+			pass
+
+		if objcScoffs is None:
+			# Older caches do not have this
+			return
+		elif objcClasslist is None:
+			self._logger.error("libobjc does not have __objc_classlist")
+			return
+
+		# Parse __objc_scoffs
+		if objcScoffs.size == 0x20:
+			# Just 4 pointers, starting with methods start and methods end
+			self._optMethodNamesAddr = self._slider.slideAddress(objcScoffs.addr)
+			_optMethodNamesEnd = self._slider.slideAddress(objcScoffs.addr + 8)
+			pass
+		elif objcScoffs.size == 0x28:
+			# First the version number and then pointers
+			verOff, ctx = self._dyldCtx.convertAddr(objcScoffs.addr)
+			version = ctx.fileCtx.readFormat("<Q", verOff)[0]
+			if version != 2:
+				self._logger.warning(f"Unknown objc opt version: {version}")
+				pass
+
+			self._optMethodNamesAddr = self._slider.slideAddress(objcScoffs.addr + 8)
+			_optMethodNamesEnd = self._slider.slideAddress(objcScoffs.addr + 16)
+			pass
+		else:
+			self._logger.error("Unable to parse objc scoffs")
+			return
+
+		optMethodNamesSize = _optMethodNamesEnd - self._optMethodNamesAddr
+
+		# Create a generator to get method lists
+		def getMethodLists() -> Generator[int, None, None]:
+			for ptrAddr in range(
+				objcClasslist.addr,
+				objcClasslist.addr + objcClasslist.size,
+				8
+			):
+				classAddr = self._slider.slideAddress(ptrAddr)
+				classDef = self._slider.slideStruct(classAddr, objc_class_t)
+				classDataDef = self._slider.slideStruct(
+					classDef.data & ~0x3,
+					objc_class_data_t
+				)
+
+				if classDataDef.baseMethods:
+					yield classDataDef.baseMethods
+				pass
+			pass
+
+		for methodListAddr in getMethodLists():
+			methodListDef = self._slider.slideStruct(methodListAddr, objc_method_list_t)
+
+			# Try to find an objc_method_small_t
+			entsize = methodListDef.getEntsize()
+			if (
+				entsize != objc_method_small_t.SIZE
+				or not methodListDef.usesRelativeMethods()
+			):
+				methodListAddr += (
+					objc_method_list_t.SIZE
+					+ (methodListDef.count * entsize)
+				)
+				continue
+
+			# Test the first method
+			methodAddr = methodListAddr + objc_method_list_t.SIZE
+			nameOff, ctx = self._dyldCtx.convertAddr(methodAddr)
+			name = ctx.fileCtx.readFormat("<i", nameOff)[0]
+
+			# TODO: Hopefully there is a better way to detect this.
+			if name >= 0 and name < optMethodNamesSize:
+				self._usesObjcRoRelativeNames = True
+				pass
+			return
+		return
+
 	def _createExtraSegment(self) -> None:
 		"""Create an extra segment to store data in.
 		"""
@@ -697,11 +594,9 @@ class _ObjCFixer(object):
 			raise _ObjCFixerError("Unable to find space for the extra ObjC segment.")
 
 		# Get a starting address for the new segment
+		leftSegOff = self._dyldCtx.convertAddr(leftSeg.vmaddr)[0]
 		newSegStartAddr = (leftSeg.vmaddr + leftSeg.vmsize + 0x1000) & ~0xFFF
-		newSegStartOff = (
-			self._dyldCtx.convertAddr(leftSeg.vmaddr) + leftSeg.vmsize
-			+ 0x1000
-		) & ~0xFFF
+		newSegStartOff = (leftSegOff + leftSeg.vmsize + 0x1000) & ~0xFFF
 
 		# adjust max gap size to account for page alignment
 		maxGapSize -= newSegStartAddr - (leftSeg.vmaddr + leftSeg.vmsize)
@@ -710,7 +605,7 @@ class _ObjCFixer(object):
 		newSegment = segment_command_64()
 		newSegment.cmd = LoadCommands.LC_SEGMENT_64
 		newSegment.cmdsize = segment_command_64.SIZE  # no sections
-		newSegment.segname = b"__EXTRA_OBJC"
+		newSegment.segname = self._extractionCtx.EXTRA_SEGMENT_NAME
 		newSegment.vmaddr = newSegStartAddr
 		newSegment.fileoff = newSegStartOff
 		newSegment.maxprot = 3  # read and write
@@ -768,6 +663,7 @@ class _ObjCFixer(object):
 					pass
 
 				elif sect.sectname == b"__objc_selrefs":
+					file = self._machoCtx.fileForAddr(sect.addr)
 					for ptrAddr in range(sect.addr, sect.addr + sect.size, 8):
 						self._statusBar.update(status="Processing Selector References")
 						selRefAddr = self._slider.slideAddress(ptrAddr)
@@ -775,8 +671,8 @@ class _ObjCFixer(object):
 						self._selRefCache[selRefAddr] = ptrAddr
 
 						newPtr = self._processString(selRefAddr)
-						self._machoCtx.writeBytes(
-							self._dyldCtx.convertAddr(ptrAddr),
+						file.writeBytes(
+							self._dyldCtx.convertAddr(ptrAddr)[0],
 							struct.pack("<Q", newPtr)
 						)
 						pass
@@ -840,8 +736,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(categoryAddr):
 			newCategoryAddr = categoryAddr
 
-			defOff = self._dyldCtx.convertAddr(categoryAddr)
-			self._machoCtx.writeBytes(defOff, categoryDef)
+			file = self._machoCtx.fileForAddr(categoryAddr)
+			defOff = self._dyldCtx.convertAddr(categoryAddr)[0]
+			file.writeBytes(defOff, categoryDef)
 			pass
 		else:
 			newCategoryAddr = self._extraDataHead
@@ -867,7 +764,7 @@ class _ObjCFixer(object):
 
 		Returns:
 			If the class if fully defined the updated address
-			of the class is returned along with False. otherwise
+			of the class is returned along with False. Otherwise
 			the original address of the class is returned, along
 			with True.
 		"""
@@ -912,8 +809,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(classAddr):
 			newClassAddr = classAddr
 
-			defOff = self._dyldCtx.convertAddr(classAddr)
-			self._machoCtx.writeBytes(defOff, classDef)
+			file = self._machoCtx.fileForAddr(classAddr)
+			defOff = self._dyldCtx.convertAddr(classAddr)[0]
+			file.writeBytes(defOff, classDef)
 			pass
 
 		else:
@@ -989,8 +887,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(classDataAddr):
 			newClassDataAddr = classDataAddr
 
-			defOff = self._dyldCtx.convertAddr(classDataAddr)
-			self._machoCtx.writeBytes(defOff, classDataDef)
+			file = self._machoCtx.fileForAddr(classDataAddr)
+			defOff = self._dyldCtx.convertAddr(classDataAddr)[0]
+			file.writeBytes(defOff, classDataDef)
 			pass
 
 		else:
@@ -1041,8 +940,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(ivarListAddr):
 			newIvarListAddr = ivarListAddr
 
-			defOff = self._dyldCtx.convertAddr(ivarListAddr)
-			self._machoCtx.writeBytes(defOff, ivarListData)
+			file = self._machoCtx.fileForAddr(ivarListAddr)
+			defOff = self._dyldCtx.convertAddr(ivarListAddr)[0]
+			file.writeBytes(defOff, ivarListData)
 			pass
 		else:
 			newIvarListAddr = self._extraDataHead
@@ -1074,8 +974,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(protoListAddr):
 			newProtoListAddr = protoListAddr
 
-			defOff = self._dyldCtx.convertAddr(protoListAddr)
-			self._machoCtx.writeBytes(defOff, protoListData)
+			file = self._machoCtx.fileForAddr(protoListAddr)
+			defOff = self._dyldCtx.convertAddr(protoListAddr)[0]
+			file.writeBytes(defOff, protoListData)
 			pass
 		else:
 			newProtoListAddr = self._extraDataHead
@@ -1143,8 +1044,9 @@ class _ObjCFixer(object):
 			newPtr = self._processString(oldPtr)
 
 			if self._machoCtx.containsAddr(protoDef.extendedMethodTypes):
-				ptrOff = self._dyldCtx.convertAddr(protoDef.extendedMethodTypes)
-				struct.pack_into("<Q", self._machoCtx.file, ptrOff, newPtr)
+				file = self._machoCtx.fileForAddr(protoDef.extendedMethodTypes)
+				ptrOff = self._dyldCtx.convertAddr(protoDef.extendedMethodTypes)[0]
+				struct.pack_into("<Q", file.file, ptrOff, newPtr)
 				pass
 			else:
 				protoDef.extendedMethodTypes = self._extraDataHead
@@ -1171,8 +1073,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(protoAddr):
 			newProtoAddr = protoAddr
 
-			defOff = self._dyldCtx.convertAddr(protoAddr)
-			self._machoCtx.writeBytes(defOff, protoData)
+			file = self._machoCtx.fileForAddr(protoAddr)
+			defOff = self._dyldCtx.convertAddr(protoAddr)[0]
+			file.writeBytes(defOff, protoData)
 			pass
 		else:
 			newProtoAddr = self._extraDataHead
@@ -1221,8 +1124,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(propertyListAddr):
 			newPropertyListAddr = propertyListAddr
 
-			defOff = self._dyldCtx.convertAddr(propertyListAddr)
-			self._machoCtx.writeBytes(defOff, propertyListData)
+			file = self._machoCtx.fileForAddr(propertyListAddr)
+			defOff = self._dyldCtx.convertAddr(propertyListAddr)[0]
+			file.writeBytes(defOff, propertyListData)
 			pass
 		else:
 			newPropertyListAddr = self._extraDataHead
@@ -1253,7 +1157,6 @@ class _ObjCFixer(object):
 		# fix relative pointers after we reserve a new address for the method list
 		# contains a list of tuples of field offsets and their target addresses
 		relativeFixups: list[tuple[int, int]] = []
-
 		for i in range(methodListDef.count):
 			methodAddr = (
 				methodListAddr
@@ -1266,11 +1169,19 @@ class _ObjCFixer(object):
 				methodOff = objc_method_list_t.SIZE + (i * entsize)
 
 				if methodDef.name:
-					nameAddr = methodAddr + methodDef.name
-					newNameAddr = self._processString(nameAddr)
-					methodDef.name = newNameAddr - methodAddr
+					if self._usesObjcRoRelativeNames:
+						baseAddr = self._optMethodNamesAddr
+						pass
+					else:
+						baseAddr = methodAddr
+						pass
 
-					relativeFixups.append((methodOff, newNameAddr))
+					nameAddr = baseAddr + methodDef.name
+					newNamePtr = self._processMethodName(nameAddr)
+
+					# make the name ptr relative to itself
+					methodDef.name = newNamePtr - methodAddr
+					relativeFixups.append((methodOff, newNamePtr))
 					pass
 
 				if methodDef.types:
@@ -1311,8 +1222,9 @@ class _ObjCFixer(object):
 		if self._machoCtx.containsAddr(methodListAddr):
 			newMethodListAddr = methodListAddr
 
-			defOff = self._dyldCtx.convertAddr(methodListAddr)
-			self._machoCtx.writeBytes(defOff, methodListData)
+			file = self._machoCtx.fileForAddr(methodListAddr)
+			defOff = self._dyldCtx.convertAddr(methodListAddr)[0]
+			file.writeBytes(defOff, methodListData)
 			pass
 		else:
 			newMethodListAddr = self._extraDataHead
@@ -1340,7 +1252,11 @@ class _ObjCFixer(object):
 		else:
 			newStringAddr = self._extraDataHead
 
-			stringData = self._dyldCtx.readString(self._dyldCtx.convertAddr(stringAddr))
+			stringOff, ctx = self._dyldCtx.convertAddr(stringAddr) or (None, None)
+			if not stringOff:
+				return None
+
+			stringData = ctx.fileCtx.readString(stringOff)
 			self._addExtraData(stringData)
 			pass
 
@@ -1357,8 +1273,8 @@ class _ObjCFixer(object):
 		else:
 			newIntAddr = self._extraDataHead
 
-			intOff = self._dyldCtx.convertAddr(intAddr)
-			intData = self._dyldCtx.getBytes(intOff, intSize)
+			intOff, ctx = self._dyldCtx.convertAddr(intAddr)
+			intData = ctx.fileCtx.getBytes(intOff, intSize)
 
 			self._addExtraData(intData)
 			pass
@@ -1366,6 +1282,26 @@ class _ObjCFixer(object):
 		self._intCache[intAddr] = newIntAddr
 		return newIntAddr
 
+	def _processMethodName(self, stringAddr: int) -> int:
+		"""Process a method name.
+
+		Returns:
+			A the address of the pointer that points
+			to the method string.
+		"""
+
+		if stringAddr in self._methodNameCache:
+			return self._methodNameCache[stringAddr]
+
+		# TODO: search selrefs first
+		newStringAddr = self._processString(stringAddr)
+		ptrAddr = self._extraDataHead
+
+		self._addExtraData(struct.pack("<Q", newStringAddr))
+
+		self._methodNameCache[stringAddr] = ptrAddr
+		return ptrAddr
+
 	def _finalizeFutureClasses(self) -> None:
 		extraSegStart = self._extraDataHead - len(self._extraData)
 
@@ -1383,119 +1319,13 @@ class _ObjCFixer(object):
 				struct.pack_into("<Q", self._extraData, ptrOffset, newAddr)
 				pass
 			else:
-				ptrOffset = self._dyldCtx.convertAddr(destPtr)
-				struct.pack_into("Q", self._machoCtx.file, ptrOffset, newAddr)
+				file = self._machoCtx.fileForAddr(destPtr)
+				ptrOffset = self._dyldCtx.convertAddr(destPtr)[0]
+				struct.pack_into("Q", file.file, ptrOffset, newAddr)
 				pass
 			pass
 		pass
 
-	def _fixSelectors_OLD(self) -> None:
-		"""Undo direct selector loading.
-
-		Changes instructions to use the selref section again.
-		"""
-
-		try:
-			textSect = self._machoCtx.segments[b"__TEXT"].sects[b"__text"]
-			pass
-		except KeyError:
-			self._logger.error("Unable to get __text section")
-			pass
-
-		self._statusBar.update(status="Fixing Selectors")
-
-		textSectAddr = textSect.addr
-		textSectOff = self._dyldCtx.convertAddr(textSectAddr)
-
-		for sectOff in range(0, textSect.size - 4, 4):
-			# Direct selector loads consist of an ADRP and ADD instruction
-			adrpOff = textSectOff + sectOff
-			if self._machoCtx.file[adrpOff + 3] & 0x9F != 0x90:
-				continue
-
-			addOff = textSectOff + sectOff + 4
-			if self._machoCtx.file[addOff + 3] != 0x91:
-				# Some times there are extra instructions in between
-				if self._machoCtx.file[addOff + 7] == 0x91:
-					addOff += 4
-					pass
-				elif self._machoCtx.file[addOff + 11] == 0x91:
-					addOff += 8
-					pass
-				else:
-					continue
-
-			adrp = self._machoCtx.readFormat(adrpOff, "<I")[0]
-			add = self._machoCtx.readFormat(addOff, "<I")[0]
-
-			# verify that the ADRP Destination register matches
-			# the ADD Base register
-			adrpDestReg = adrp & 0x1F
-			addBaseReg = (add >> 5) & 0x1F
-			if adrpDestReg != addBaseReg:
-				continue
-
-			# get the selector address
-			adrpAddr = textSectAddr + sectOff
-
-			# ADRP
-			immlo = (adrp & 0x60000000) >> 29
-			immhi = (adrp & 0xFFFFE0) >> 3
-			imm = (immhi | immlo) << 12
-			imm = stub_fixer.Arm64Utilities.signExtend(imm, 33)
-
-			adrpResult = (adrpAddr & ~0xFFF) + imm
-
-			# ADD
-			imm = (add & 0x3FFC00) >> 10
-			loadTarget = adrpResult + imm
-
-			if self._machoCtx.containsAddr(loadTarget):
-				continue
-
-			if loadTarget not in self._selRefCache:
-				# There are some files that access strings that do not have
-				# a selector reference. Pull in the string and repoint the
-				# ADRP and ADD to it.
-				stringAddr = self._processString(loadTarget)
-
-				adrpDelta = (stringAddr & -4096) - (adrpAddr & -4096)
-				immhi = (adrpDelta >> 9) & (0x00FFFFE0)
-				immlo = (adrpDelta << 17) & (0x60000000)
-				newAdrp = (0x90000000) | immlo | immhi | adrpDestReg
-
-				addOff = stringAddr - (stringAddr & -4096)
-				imm12 = (addOff << 10) & 0x3FFC00
-				addDestReg = add & 0x1F
-				newAdd = 0x91000000 | imm12 | (addBaseReg << 5) | addDestReg
-
-				self._machoCtx.writeBytes(adrpOff, struct.pack("<I", newAdrp))
-				self._machoCtx.writeBytes(addOff, struct.pack("<I", newAdd))
-
-				self._statusBar.update(status="Fixing Selectors")
-				continue
-
-			selRefPtr = self._selRefCache[loadTarget]
-
-			# make new adrp and ldr instructions
-			adrpDelta = (selRefPtr & -4096) - (adrpAddr & -4096)
-			immhi = (adrpDelta >> 9) & (0x00FFFFE0)
-			immlo = (adrpDelta << 17) & (0x60000000)
-			newAdrp = 0x90000000 | immlo | immhi | adrpDestReg
-
-			ldrOffset = selRefPtr - (selRefPtr & -4096)
-			imm12 = (ldrOffset << 7) & 0x3FFC00
-			ldrDestReg = add & 0x1F
-			newLdr = 0xF9400000 | imm12 | (addBaseReg << 5) | ldrDestReg
-
-			# write to new instructions
-			self._machoCtx.writeBytes(adrpOff, struct.pack("<I", newAdrp))
-			self._machoCtx.writeBytes(addOff, struct.pack("<I", newLdr))
-
-			self._statusBar.update(status="Fixing Selectors")
-			pass
-		pass
-
 	def _checkSpaceConstraints(self) -> None:
 		"""Check if we have enough space to add the new segment.
 		"""
@@ -1559,25 +1389,27 @@ class _ObjCFixer(object):
 			if cmd in commandsToRemove:
 				continue
 
-			loadCommandsData.extend(self._machoCtx.getBytes(readHead, cmd.cmdsize))
+			loadCommandsData.extend(
+				self._machoCtx.fileCtx.getBytes(readHead, cmd.cmdsize)
+			)
 			readHead += cmd.cmdsize
 			pass
 
 		self._machoCtx.header.ncmds -= len(commandsToRemove)
 		self._machoCtx.header.sizeofcmds = len(loadCommandsData)
-		self._machoCtx.writeBytes(
+		self._machoCtx.fileCtx.writeBytes(
 			self._machoCtx.fileOffset + mach_header_64.SIZE,
 			loadCommandsData
 		)
 
-		self._machoCtx = MachOContext(self._machoCtx.file, self._machoCtx.fileOffset)
-		self._extractionCtx.machoCtx = self._machoCtx
+		self._machoCtx.reloadLoadCommands()
 		pass
 
 	def _addExtraDataSeg(self) -> None:
-		# update the size on the new segment
-		self._extraSegment.vmsize = len(self._extraData)
-		self._extraSegment.filesize = len(self._extraData)
+		# update the size on the new segment and mappings
+		extraDataLen = len(self._extraData)
+		self._extraSegment.vmsize = extraDataLen
+		self._extraSegment.filesize = extraDataLen
 
 		# insert the segment command right before the linkedit
 		moveStart = self._machoCtx.segments[b"__LINKEDIT"].seg._fileOff_
@@ -1587,25 +1419,22 @@ class _ObjCFixer(object):
 			+ self._machoCtx.header.sizeofcmds
 			- moveStart
 		)
-		self._machoCtx.file.move(
+		self._machoCtx.fileCtx.file.move(
 			moveStart + segment_command_64.SIZE,
 			moveStart,
 			bytesToMove
 		)
 
-		# add the new data and the segment command
-		self._machoCtx.writeBytes(self._extraSegment.fileoff, self._extraData)
-		self._machoCtx.writeBytes(moveStart, self._extraSegment)
+		self._machoCtx.fileCtx.writeBytes(moveStart, self._extraSegment)
+
+		# update the extraction context
+		self._extractionCtx.extraSegmentData = self._extraData
 
 		# update the header
 		self._machoCtx.header.ncmds += 1
 		self._machoCtx.header.sizeofcmds += segment_command_64.SIZE
 
-		# recreate the macho context to reflect the new segment
-		self._extractionCtx.machoCtx = MachOContext(
-			self._machoCtx.file,
-			self._machoCtx.fileOffset
-		)
+		self._machoCtx.reloadLoadCommands()
 		pass
 	pass
 
