@@ -53,7 +53,7 @@ class _Symbolizer(object):
 		# create a map of image paths and their addresses
 		self._images: dict[bytes, int] = {}
 		for image in self._dyldCtx.images:
-			imagePath = self._dyldCtx.readString(image.pathFileOffset)
+			imagePath = self._dyldCtx.fileCtx.readString(image.pathFileOffset)
 			self._images[imagePath] = image.address
 			pass
 
@@ -99,7 +99,7 @@ class _Symbolizer(object):
 		# get an initial list of dependencies
 		if dylibs := self._machoCtx.getLoadCommand(DEP_LCS, multiple=True):
 			for dylib in dylibs:
-				if depInfo := self._getDepInfo(dylib):
+				if depInfo := self._getDepInfo(dylib, self._machoCtx):
 					depsQueue.append(depInfo)
 			pass
 
@@ -123,8 +123,8 @@ class _Symbolizer(object):
 			if dylibs := depInfo.context.getLoadCommand(DEP_LCS, multiple=True):
 				for dylib in dylibs:
 					if dylib.cmd == LoadCommands.LC_REEXPORT_DYLIB:
-						if depInfo := self._getDepInfo(dylib):
-							depsQueue.append(depInfo)
+						if info := self._getDepInfo(dylib, depInfo.context):
+							depsQueue.append(info)
 				pass
 
 			# check for any ReExport exports
@@ -137,8 +137,8 @@ class _Symbolizer(object):
 
 			for ordinal in reExportOrdinals:
 				dylib = dylibs[ordinal - 1]
-				if depInfo := self._getDepInfo(dylib):
-					depsQueue.append(depInfo)
+				if info := self._getDepInfo(dylib, depInfo.context):
+					depsQueue.append(info)
 				pass
 			pass
 
@@ -161,19 +161,25 @@ class _Symbolizer(object):
 				self._logger.warning(f"No root export for ReExport with symbol {name}")
 		pass
 
-	def _getDepInfo(self, dylib: dylib_command) -> _DependencyInfo:
+	def _getDepInfo(
+		self,
+		dylib: dylib_command,
+		context: MachOContext
+	) -> _DependencyInfo:
 		"""Given a dylib command, get dependency info.
 		"""
 
 		dylibPathOff = dylib._fileOff_ + dylib.dylib.name.offset
-		dylibPath = self._dyldCtx.readString(dylibPathOff)
+		dylibPath = context.fileCtx.readString(dylibPathOff)
 		if dylibPath not in self._images:
 			self._logger.warning(f"Unable to find dependency: {dylibPath}")
 			return None
 
 		imageAddr = self._images[dylibPath]
-		imageOff = self._dyldCtx.convertAddr(imageAddr)
-		context = MachOContext(self._dyldCtx.file, imageOff)
+		imageOff, dyldCtx = self._dyldCtx.convertAddr(imageAddr)
+
+		# Since we're not editing the dependencies, this should be fine.
+		context = MachOContext(dyldCtx.fileCtx, imageOff)
 		return _DependencyInfo(dylibPath, imageAddr, context)
 
 	def _readDepExports(
@@ -201,9 +207,13 @@ class _Symbolizer(object):
 			# Some images like UIKit don't have exports
 			return []
 
+		linkeditFile = self._dyldCtx.convertAddr(
+			depInfo.context.segments[b"__LINKEDIT"].seg.vmaddr
+		)[1].fileCtx.file
+
 		try:
 			depExports = dyld_trie.ReadExports(
-				depInfo.context.file,
+				linkeditFile,
 				exportOff,
 				exportSize,
 			)
@@ -250,21 +260,19 @@ class _Symbolizer(object):
 			self._logger.warning("Unable to find LC_SYMTAB.")
 			return
 
-		dysymtab: dysymtab_command = self._machoCtx.getLoadCommand(
-			(LoadCommands.LC_DYSYMTAB,)
+		linkeditFile = self._machoCtx.fileForAddr(
+			self._machoCtx.segments[b"__LINKEDIT"].seg.vmaddr
 		)
-		if not dysymtab:
-			self._logger.warning("Unable to find LC_DYSYMTAB.")
 
 		for i in range(symtab.nsyms):
 			self._statusBar.update()
 
 			# Get the symbol and its address
 			entryOff = symtab.symoff + (i * nlist_64.SIZE)
-			symbolEntry = nlist_64(self._machoCtx.file, entryOff)
+			symbolEntry = nlist_64(linkeditFile.file, entryOff)
 
 			symbolAddr = symbolEntry.n_value
-			symbol = self._machoCtx.readString(symtab.stroff + symbolEntry.n_strx)
+			symbol = linkeditFile.readString(symtab.stroff + symbolEntry.n_strx)
 
 			if symbolAddr == 0:
 				continue
@@ -450,10 +458,11 @@ class Arm64Utilities(object):
 			If unable to get the bind data, return None.
 		"""
 
-		if not (helperOff := self._dyldCtx.convertAddr(address)):
+		helperOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if helperOff is None:
 			return None
 
-		ldr, b, data = self._dyldCtx.readFormat(helperOff, "<III")
+		ldr, b, data = ctx.fileCtx.readFormat("<III", helperOff)
 
 		# verify
 		if (
@@ -516,11 +525,12 @@ class Arm64Utilities(object):
 
 		SEARCH_LIMIT = 0xC8
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
 		# test stp and mov
-		stp, mov = self._dyldCtx.readFormat(stubOff, "<II")
+		stp, mov = ctx.fileCtx.readFormat("<II", stubOff)
 		if (
 			(stp & 0x7FC00000) != 0x29800000
 			or (mov & 0x7F3FFC00) != 0x11000000
@@ -528,7 +538,7 @@ class Arm64Utilities(object):
 			return None
 
 		# Find the branch instruction
-		dataSource = self._dyldCtx.file
+		dataSource = ctx.fileCtx.file
 		branchInstrOff = None
 		for instrOff in range(stubOff, stubOff + SEARCH_LIMIT, 4):
 			# (instr & 0xFE9FF000) == 0xD61F0000
@@ -557,8 +567,8 @@ class Arm64Utilities(object):
 			return None
 
 		# Test if there is a stp before the bl and a ldp before the braaz
-		adrp = self._dyldCtx.readFormat(blInstrOff + 4, "<I")[0]
-		ldp = self._dyldCtx.readFormat(branchInstrOff - 4, "<I")[0]
+		adrp = ctx.fileCtx.readFormat("<I", blInstrOff + 4)[0]
+		ldp = ctx.fileCtx.readFormat("<I", branchInstrOff - 4)[0]
 		if (
 			(adrp & 0x9F00001F) != 0x90000010
 			or (ldp & 0x7FC00000) != 0x28C00000
@@ -566,7 +576,7 @@ class Arm64Utilities(object):
 			return None
 
 		# Hopefully it's a resolver...
-		imm = (self._dyldCtx.readFormat(blInstrOff, "<I")[0] & 0x3FFFFFF) << 2
+		imm = (ctx.fileCtx.readFormat("<I", blInstrOff)[0] & 0x3FFFFFF) << 2
 		imm = self.signExtend(imm, 28)
 		blResult = address + (blInstrOff - stubOff) + imm
 
@@ -609,10 +619,11 @@ class Arm64Utilities(object):
 			be determined.
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, ldr, br = self._dyldCtx.readFormat(stubOff, "<III")
+		adrp, ldr, br = ctx.fileCtx.readFormat("<III", stubOff)
 
 		# verify
 		if (
@@ -645,12 +656,13 @@ class Arm64Utilities(object):
 			not be determined.
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, add, ldr, braa = self._dyldCtx.readFormat(
-			stubOff,
-			"<IIII"
+		adrp, add, ldr, braa = ctx.fileCtx.readFormat(
+			"<IIII",
+			stubOff
 		)
 
 		# verify
@@ -684,10 +696,11 @@ class Arm64Utilities(object):
 		BR x16
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, ldr, br = self._dyldCtx.readFormat(stubOff, "<III")
+		adrp, ldr, br = ctx.fileCtx.readFormat("<III", stubOff)
 
 		# verify
 		if (
@@ -716,10 +729,11 @@ class Arm64Utilities(object):
 		BR x16
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, add, br = self._dyldCtx.readFormat(stubOff, "<III")
+		adrp, add, br = ctx.fileCtx.readFormat("<III", stubOff)
 
 		# verify
 		if (
@@ -748,10 +762,11 @@ class Arm64Utilities(object):
 		11 0a 1f d7  braa  	x16=>__auth_stubs::_CCRandomCopyBytes,x17
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, add, ldr, braa = self._dyldCtx.readFormat(stubOff, "<IIII")
+		adrp, add, ldr, braa = ctx.fileCtx.readFormat("<IIII", stubOff)
 
 		# verify
 		if (
@@ -787,10 +802,11 @@ class Arm64Utilities(object):
 		1bfcb5d2c 20 00 20 d4  trap
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, add, br, trap = self._dyldCtx.readFormat(stubOff, "<IIII")
+		adrp, add, br, trap = ctx.fileCtx.readFormat("<IIII", stubOff)
 
 		# verify
 		if (
@@ -819,10 +835,11 @@ class Arm64Utilities(object):
 		1f 0a 1f d6  braaz 	x16=>FUN_195bee070
 		"""
 
-		if not (stubOff := self._dyldCtx.convertAddr(address)):
+		stubOff, ctx = self._dyldCtx.convertAddr(address) or (None, None)
+		if stubOff is None:
 			return None
 
-		adrp, ldr, braaz = self._dyldCtx.readFormat(stubOff, "<III")
+		adrp, ldr, braaz = ctx.fileCtx.readFormat("<III", stubOff)
 
 		# verify
 		if (
@@ -1024,6 +1041,11 @@ class _StubFixer(object):
 		dyldInfo: dyld_info_command = self._machoCtx.getLoadCommand(
 			(LoadCommands.LC_DYLD_INFO, LoadCommands.LC_DYLD_INFO_ONLY)
 		)
+
+		linkeditFile = self._machoCtx.fileForAddr(
+			self._machoCtx.segments[b"__LINKEDIT"].seg.vmaddr
+		)
+
 		if dyldInfo:
 			records: List[_BindRecord] = []
 			try:
@@ -1031,7 +1053,7 @@ class _StubFixer(object):
 					# usually contains records for c++ symbols like "new"
 					records.extend(
 						_bindReader(
-							self._machoCtx,
+							linkeditFile,
 							dyldInfo.weak_bind_off,
 							dyldInfo.weak_bind_size
 						)
@@ -1041,7 +1063,7 @@ class _StubFixer(object):
 				if dyldInfo.lazy_bind_off:
 					records.extend(
 						_bindReader(
-							self._machoCtx,
+							linkeditFile,
 							dyldInfo.lazy_bind_off,
 							dyldInfo.lazy_bind_size
 						)
@@ -1098,9 +1120,9 @@ class _StubFixer(object):
 							continue
 
 						# Try to symbolize though indirect symbol entries
-						symbolIndex = self._machoCtx.readFormat(
-							self._dysymtab.indirectsymoff + ((sect.reserved1 + i) * 4),
-							"<I"
+						symbolIndex = linkeditFile.readFormat(
+							"<I",
+							self._dysymtab.indirectsymoff + ((sect.reserved1 + i) * 4)
 						)[0]
 						if (
 							symbolIndex != 0
@@ -1109,10 +1131,10 @@ class _StubFixer(object):
 							and symbolIndex != (INDIRECT_SYMBOL_ABS | INDIRECT_SYMBOL_LOCAL)
 						):
 							symbolEntry = nlist_64(
-								self._machoCtx.file,
+								linkeditFile.file,
 								self._symtab.symoff + (symbolIndex * nlist_64.SIZE)
 							)
-							symbol = self._machoCtx.readString(
+							symbol = linkeditFile.readString(
 								self._symtab.stroff + symbolEntry.n_strx
 							)
 
@@ -1157,6 +1179,10 @@ class _StubFixer(object):
 		if not dyldInfo:
 			return
 
+		linkeditFile = self._machoCtx.fileForAddr(
+			self._machoCtx.segments[b"__LINKEDIT"].seg.vmaddr
+		)
+
 		# the stub helper section has the stub binder in
 		# beginning, skip it.
 		helperAddr = helperSect.addr + STUB_BINDER_SIZE
@@ -1168,7 +1194,7 @@ class _StubFixer(object):
 			if (bindOff := self._arm64Utils.getStubHelperData(helperAddr)) is not None:
 				record = next(
 					_bindReader(
-						self._machoCtx,
+						linkeditFile,
 						dyldInfo.lazy_bind_off + bindOff,
 						dyldInfo.lazy_bind_size,
 					),
@@ -1186,11 +1212,12 @@ class _StubFixer(object):
 					continue
 
 				# repoint the bind pointer to the stub helper
-				bindPtrOff = self._machoCtx.segmentsI[record.segment].seg.fileoff
-				bindPtrOff += record.offset
+				bindPtrAddr = self._machoCtx.segmentsI[record.segment].seg.vmaddr
+				bindPtrOff = self._dyldCtx.convertAddr(bindPtrAddr)[0] + record.offset
+				ctx = self._machoCtx.fileForAddr(bindPtrAddr)
 
 				newBindPtr = struct.pack("<Q", helperAddr)
-				self._machoCtx.writeBytes(bindPtrOff, newBindPtr)
+				ctx.writeBytes(bindPtrOff, newBindPtr)
 				helperAddr += REG_HELPER_SIZE
 				continue
 
@@ -1224,6 +1251,16 @@ class _StubFixer(object):
 				stubMap[stubName] = [stubAddr]
 			pass
 
+		linkeditFile = self._machoCtx.fileForAddr(
+			self._machoCtx.segments[b"__LINKEDIT"].seg.vmaddr
+		)
+
+		textFile = self._machoCtx.fileForAddr(
+			self._machoCtx.segments[b"__TEXT"].seg.vmaddr
+		)
+
+		symbolPtrFile = None
+
 		for segment in self._machoCtx.segmentsI:
 			for sect in segment.sectsI:
 				if sect.flags & SECTION_TYPE == S_SYMBOL_STUBS:
@@ -1236,9 +1273,9 @@ class _StubFixer(object):
 						stubNames = None
 
 						# Try to symbolize though indirect symbol entries
-						symbolIndex = self._machoCtx.readFormat(
-							self._dysymtab.indirectsymoff + ((sect.reserved1 + i) * 4),
-							"<I"
+						symbolIndex = linkeditFile.readFormat(
+							"<I",
+							self._dysymtab.indirectsymoff + ((sect.reserved1 + i) * 4)
 						)[0]
 
 						if (
@@ -1248,11 +1285,11 @@ class _StubFixer(object):
 							and symbolIndex != (INDIRECT_SYMBOL_ABS | INDIRECT_SYMBOL_LOCAL)
 						):
 							symbolEntry = nlist_64(
-								self._machoCtx.file,
+								linkeditFile.file,
 								self._symtab.symoff + (symbolIndex * nlist_64.SIZE)
 							)
 							stubNames = [
-								self._machoCtx.readString(self._symtab.stroff + symbolEntry.n_strx)
+								linkeditFile.readString(self._symtab.stroff + symbolEntry.n_strx)
 							]
 							pass
 
@@ -1309,32 +1346,32 @@ class _StubFixer(object):
 							elif stubFormat == _StubFormat.StubOptimized:
 								# only need to relink stub
 								newStub = self._arm64Utils.generateStubNormal(stubAddr, symPtrAddr)
-								self._machoCtx.writeBytes(
-									self._dyldCtx.convertAddr(stubAddr),
-									newStub
-								)
+								stubOff = self._dyldCtx.convertAddr(stubAddr)[0]
+								textFile.writeBytes(stubOff, newStub)
 								continue
 
 							elif stubFormat == _StubFormat.AuthStubNormal:
 								# only need to relink symbol pointer
-								self._machoCtx.writeBytes(
-									self._dyldCtx.convertAddr(symPtrAddr),
-									struct.pack("<Q", stubAddr)
-								)
+								symPtrOff = self._dyldCtx.convertAddr(symPtrAddr)[0]
+
+								if not symbolPtrFile:
+									symbolPtrFile = self._machoCtx.fileForAddr(symPtrAddr)
+									pass
+
+								symbolPtrFile.writeBytes(symPtrOff, struct.pack("<Q", stubAddr))
 								continue
 
 							elif stubFormat == _StubFormat.AuthStubOptimized:
 								# need to relink both the stub and the symbol pointer
-								self._machoCtx.writeBytes(
-									self._dyldCtx.convertAddr(symPtrAddr),
-									struct.pack("<Q", stubAddr)
-								)
+								symPtrOff = self._dyldCtx.convertAddr(symPtrAddr)[0]
+								if not symbolPtrFile:
+									symbolPtrFile = self._machoCtx.fileForAddr(symPtrAddr)
+									pass
+								symbolPtrFile.writeBytes(symPtrOff, struct.pack("<Q", stubAddr))
 
 								newStub = self._arm64Utils.generateAuthStubNormal(stubAddr, symPtrAddr)
-								self._machoCtx.writeBytes(
-									self._dyldCtx.convertAddr(stubAddr),
-									newStub
-								)
+								stubOff, ctx = self._dyldCtx.convertAddr(stubAddr)
+								textFile.writeBytes(stubOff, newStub)
 								continue
 
 							elif stubFormat == _StubFormat.AuthStubResolver:
@@ -1372,7 +1409,8 @@ class _StubFixer(object):
 		textAddr = textSect.addr
 		# Section offsets by section_64.offset are sometimes
 		# inaccurate, like in libcrypto.dylib
-		textOff = self._dyldCtx.convertAddr(textAddr)
+		textOff = self._dyldCtx.convertAddr(textAddr)[0]
+		textFile = self._machoCtx.fileForAddr(textAddr)
 
 		for sectOff in range(0, textSect.size, 4):
 			# We are only looking for bl and b instructions only.
@@ -1380,7 +1418,7 @@ class _StubFixer(object):
 			# most byte. By only looking at the top byte, we can
 			# save a lot of time.
 			instrOff = textOff + sectOff
-			instrTop = self._machoCtx.file[instrOff + 3] & 0xFC
+			instrTop = textFile.file[instrOff + 3] & 0xFC
 
 			if (
 				instrTop != 0x94  # bl
@@ -1389,7 +1427,7 @@ class _StubFixer(object):
 				continue
 
 			# get the target of the branch
-			brInstr = self._machoCtx.readFormat(instrOff, "<I")[0]
+			brInstr = textFile.readFormat("<I", instrOff)[0]
 			imm26 = brInstr & 0x3FFFFFF
 			brOff = self._arm64Utils.signExtend(imm26 << 2, 28)
 
@@ -1406,7 +1444,7 @@ class _StubFixer(object):
 				# Sometimes there are bytes of data in the text section
 				# that match the bl and b filter, these seem to follow a
 				# BR or other branch, skip these.
-				lastInstTop = self._machoCtx.file[instrOff + 3] & 0xFC
+				lastInstTop = textFile.file[instrOff + 3] & 0xFC
 				if (
 					lastInstTop == 0x94  # bl
 					or lastInstTop == 0x14  # b
@@ -1420,7 +1458,7 @@ class _StubFixer(object):
 			stubSymbol = next((sym for sym in funcSymbols if sym in stubMap), None)
 			if not stubSymbol:
 				# Same as above
-				lastInstTop = self._machoCtx.file[instrOff + 3] & 0xFC
+				lastInstTop = textFile.file[instrOff + 3] & 0xFC
 				if (
 					lastInstTop == 0x94  # bl
 					or lastInstTop == 0x14  # b
@@ -1435,7 +1473,7 @@ class _StubFixer(object):
 			stubAddr = stubMap[stubSymbol][0]
 			imm26 = (stubAddr - brAddr) >> 2
 			brInstr = (brInstr & 0xFC000000) | imm26
-			struct.pack_into("<I", self._machoCtx.file, instrOff, brInstr)
+			struct.pack_into("<I", textFile.file, instrOff, brInstr)
 
 			self._statusBar.update(status="Fixing Callsites")
 			pass
@@ -1459,6 +1497,10 @@ class _StubFixer(object):
 
 		self._statusBar.update(status="Fixing Indirect Symbols")
 
+		linkeditFile = self._machoCtx.fileForAddr(
+			self._machoCtx.segments[b"__LINKEDIT"].seg.vmaddr
+		)
+
 		currentSymbolIndex = self._dysymtab.iundefsym + self._dysymtab.nundefsym
 		currentStringIndex = self._symtab.strsize
 
@@ -1476,7 +1518,7 @@ class _StubFixer(object):
 						self._statusBar.update()
 
 						entryOffset = self._dysymtab.indirectsymoff + (i * 4)
-						entry = self._machoCtx.readFormat(entryOffset, "<I")[0]
+						entry = linkeditFile.readFormat("<I", entryOffset)[0]
 
 						if entry != 0:
 							continue
@@ -1499,7 +1541,7 @@ class _StubFixer(object):
 						currentStringIndex += len(stubSymbol)
 
 						# update the indirect entry and add it
-						self._machoCtx.writeBytes(
+						linkeditFile.writeBytes(
 							entryOffset,
 							struct.pack("<I", currentSymbolIndex)
 						)
@@ -1520,7 +1562,7 @@ class _StubFixer(object):
 						self._statusBar.update()
 
 						entryOffset = self._dysymtab.indirectsymoff + (i * 4)
-						entry = self._machoCtx.readFormat(entryOffset, "<I")[0]
+						entry = linkeditFile.readFormat("<I", entryOffset)[0]
 
 						if entry != 0:
 							continue
@@ -1543,7 +1585,7 @@ class _StubFixer(object):
 						currentStringIndex += len(ptrSymbol)
 
 						# update the indirect entry and add it
-						self._machoCtx.writeBytes(
+						linkeditFile.writeBytes(
 							entryOffset,
 							struct.pack("<I", currentSymbolIndex)
 						)
@@ -1564,7 +1606,7 @@ class _StubFixer(object):
 						self._statusBar.update()
 
 						entryOffset = self._dysymtab.indirectsymoff + (i * 4)
-						entry = self._machoCtx.readFormat(entryOffset, "<I")[0]
+						entry = linkeditFile.readFormat("<I", entryOffset)[0]
 
 						if entry != 0:
 							continue
@@ -1580,7 +1622,7 @@ class _StubFixer(object):
 						self._statusBar.update()
 
 						entryOffset = self._dysymtab.indirectsymoff + (i * 4)
-						entry = self._machoCtx.readFormat(entryOffset, "<I")[0]
+						entry = linkeditFile.readFormat("<I", entryOffset)[0]
 
 						if entry != 0:
 							continue
@@ -1604,11 +1646,11 @@ class _StubFixer(object):
 		self._statusBar.update()
 
 		# add the new data and update the load commands
-		self._machoCtx.writeBytes(
+		linkeditFile.writeBytes(
 			self._symtab.symoff + (self._symtab.nsyms * nlist_64.SIZE),
 			newSymbols
 		)
-		self._machoCtx.writeBytes(
+		linkeditFile.writeBytes(
 			self._symtab.stroff + self._symtab.strsize,
 			newStrings
 		)
